"""
Implementation of the Hungarian (Munkres) Algorithm using Python and NumPy
References: http://www.ams.jhu.edu/~castello/362/Handouts/hungarian.pdf
        http://weber.ucsd.edu/~vcrawfor/hungar.pdf
        http://en.wikipedia.org/wiki/Hungarian_algorithm
        http://www.public.iastate.edu/~ddoty/HungarianAlgorithm.html
        http://www.clapper.org/software/python/munkres/
"""

# Module Information.
__version__ = "1.1.1"
__author__ = "Thom Dedecko"
__url__ = "http://github.com/tdedecko/hungarian-algorithm"
__copyright__ = "(c) 2010 Thom Dedecko"
__license__ = "MIT License"


class HungarianError(Exception):
    pass


# Import numpy. Error if fails
try:
    import numpy as np
except ImportError:
    raise HungarianError("NumPy is not installed.")


class Hungarian:
    """
    Implementation of the Hungarian (Munkres) Algorithm using np.

    Usage:
        hungarian = Hungarian(cost_matrix)
        hungarian.calculate()
    or
        hungarian = Hungarian()
        hungarian.calculate(cost_matrix)

    Handle Profit matrix:
        hungarian = Hungarian(profit_matrix, is_profit_matrix=True)
    or
        cost_matrix = Hungarian.make_cost_matrix(profit_matrix)

    The matrix will be automatically padded if it is not square.
    For that numpy's resize function is used, which automatically adds 0's to any row/column that is added

    Get results and total potential after calculation:
        hungarian.get_results()
        hungarian.get_total_potential()
    """

    def __init__(self, input_matrix=None, is_profit_matrix=False):
        """
        input_matrix is a List of Lists.
        input_matrix is assumed to be a cost matrix unless is_profit_matrix is True.
        """
        if input_matrix is not None:
            # Save input
            my_matrix = np.array(input_matrix)
            self._input_matrix = np.array(input_matrix)
            self._maxColumn = my_matrix.shape[1]
            self._maxRow = my_matrix.shape[0]

            # Adds 0s if any columns/rows are added. Otherwise stays unaltered
            matrix_size = max(self._maxColumn, self._maxRow)
            pad_columns = matrix_size - self._maxRow
            pad_rows = matrix_size - self._maxColumn
            my_matrix = np.pad(
                my_matrix,
                ((0, pad_columns), (0, pad_rows)),
                "constant",
                constant_values=(0),
            )

            # Convert matrix to profit matrix if necessary
            if is_profit_matrix:
                my_matrix = self.make_cost_matrix(my_matrix)

            self._cost_matrix = my_matrix
            self._size = len(my_matrix)
            self._shape = my_matrix.shape

            # Results from algorithm.
            self._results = []
            self._totalPotential = 0
        else:
            self._cost_matrix = None

    def get_results(self):
        """Get results after calculation."""
        return self._results

    def get_total_potential(self):
        """Returns expected value after calculation."""
        return self._totalPotential

    def calculate(self, input_matrix=None, is_profit_matrix=False):
        """
        Implementation of the Hungarian (Munkres) Algorithm.

        input_matrix is a List of Lists.
        input_matrix is assumed to be a cost matrix unless is_profit_matrix is True.
        """
        # Handle invalid and new matrix inputs.
        if input_matrix is None and self._cost_matrix is None:
            raise HungarianError("Invalid input")
        elif input_matrix is not None:
            self.__init__(input_matrix, is_profit_matrix)

        result_matrix = self._cost_matrix.copy()

        # Step 1: Subtract row mins from each row.
        for index, row in enumerate(result_matrix):
            result_matrix[index] -= row.min()

        # Step 2: Subtract column mins from each column.
        for index, column in enumerate(result_matrix.T):
            result_matrix[:, index] -= column.min()

        # Step 3: Use minimum number of lines to cover all zeros in the matrix.
        # If the total covered rows+columns is not equal to the matrix size then adjust matrix and repeat.
        total_covered = 0
        while total_covered < self._size:
            # Find minimum number of lines to cover all zeros in the matrix and find total covered rows and columns.
            cover_zeros = CoverZeros(result_matrix)
            covered_rows = cover_zeros.get_covered_rows()
            covered_columns = cover_zeros.get_covered_columns()
            total_covered = len(covered_rows) + len(covered_columns)

            # if the total covered rows+columns is not equal to the matrix size then adjust it by min uncovered num (m).
            if total_covered < self._size:
                result_matrix = self._adjust_matrix_by_min_uncovered_num(
                    result_matrix, covered_rows, covered_columns
                )

        # Step 4: Starting with the top row, work your way downwards as you make assignments.
        # Find single zeros in rows or columns.
        # Add them to final result and remove them and their associated row/column from the matrix.
        expected_results = min(self._maxColumn, self._maxRow)
        zero_locations = result_matrix == 0
        while len(self._results) != expected_results:

            # If number of zeros in the matrix is zero before finding all the results then an error has occurred.
            if not zero_locations.any():
                raise HungarianError("Unable to find results. Algorithm has failed.")

            # Find results and mark rows and columns for deletion
            matched_rows, matched_columns = self.__find_matches(zero_locations)

            # Make arbitrary selection
            total_matched = len(matched_rows) + len(matched_columns)
            if total_matched == 0:
                matched_rows, matched_columns = self.select_arbitrary_match(
                    zero_locations
                )

            # Delete rows and columns
            for row in matched_rows:
                zero_locations[row] = False
            for column in matched_columns:
                zero_locations[:, column] = False

            # Save Results
            self.__set_results(zip(matched_rows, matched_columns))

        # Calculate total potential
        value = 0
        for row, column in self._results:
            value += self._input_matrix[row, column]
        self._totalPotential = value

    @staticmethod
    def make_cost_matrix(profit_matrix):
        """
        Converts a profit matrix into a cost matrix.
        Expects NumPy objects as input.
        """
        # subtract profit matrix from a matrix made of the max value of the profit matrix
        matrix_shape = profit_matrix.shape
        offset_matrix = np.ones(matrix_shape, dtype=int) * profit_matrix.max()
        cost_matrix = offset_matrix - profit_matrix
        return cost_matrix

    def _adjust_matrix_by_min_uncovered_num(
        self, result_matrix, covered_rows, covered_columns
    ):
        """Subtract m from every uncovered number and add m to every element covered with two lines."""
        # Calculate minimum uncovered number (m)
        elements = []
        for row_index, row in enumerate(result_matrix):
            if row_index not in covered_rows:
                for index, element in enumerate(row):
                    if index not in covered_columns:
                        elements.append(element)
        min_uncovered_num = min(elements)

        # Add m to every covered element
        adjusted_matrix = result_matrix
        for row in covered_rows:
            adjusted_matrix[row] += min_uncovered_num
        for column in covered_columns:
            adjusted_matrix[:, column] += min_uncovered_num

        # Subtract m from every element
        m_matrix = np.ones(self._shape, dtype=int) * min_uncovered_num
        adjusted_matrix -= m_matrix

        return adjusted_matrix

    def __find_matches(self, zero_locations):
        """Returns rows and columns with matches in them."""
        marked_rows = np.array([], dtype=int)
        marked_columns = np.array([], dtype=int)

        # Mark rows and columns with matches
        # Iterate over rows
        for index, row in enumerate(zero_locations):
            row_index = np.array([index])
            if np.sum(row) == 1:
                (column_index,) = np.where(row)
                marked_rows, marked_columns = self.__mark_rows_and_columns(
                    marked_rows, marked_columns, row_index, column_index
                )

        # Iterate over columns
        for index, column in enumerate(zero_locations.T):
            column_index = np.array([index])
            if np.sum(column) == 1:
                (row_index,) = np.where(column)
                marked_rows, marked_columns = self.__mark_rows_and_columns(
                    marked_rows, marked_columns, row_index, column_index
                )

        return marked_rows, marked_columns

    @staticmethod
    def __mark_rows_and_columns(marked_rows, marked_columns, row_index, column_index):
        """Check if column or row is marked. If not marked then mark it."""
        new_marked_rows = marked_rows
        new_marked_columns = marked_columns
        if (
            not (marked_rows == row_index).any()
            and not (marked_columns == column_index).any()
        ):
            new_marked_rows = np.insert(marked_rows, len(marked_rows), row_index)
            new_marked_columns = np.insert(
                marked_columns, len(marked_columns), column_index
            )
        return new_marked_rows, new_marked_columns

    @staticmethod
    def select_arbitrary_match(zero_locations):
        """Selects row column combination with minimum number of zeros in it."""
        # Count number of zeros in row and column combinations
        rows, columns = np.where(zero_locations)
        zero_count = []
        for index, row in enumerate(rows):
            total_zeros = np.sum(zero_locations[row]) + np.sum(
                zero_locations[:, columns[index]]
            )
            zero_count.append(total_zeros)

        # Get the row column combination with the minimum number of zeros.
        indices = zero_count.index(min(zero_count))
        row = np.array([rows[indices]])
        column = np.array([columns[indices]])

        return row, column

    def __set_results(self, result_lists):
        """Set results during calculation."""
        # Check if results values are out of bound from input matrix (because of matrix being padded).
        # Add results to results list.
        for result in result_lists:
            row, column = result
            if row < self._maxRow and column < self._maxColumn:
                new_result = (int(row), int(column))
                self._results.append(new_result)


class CoverZeros:
    """
    Use minimum number of lines to cover all zeros in the matrix.
    Algorithm based on: http://weber.ucsd.edu/~vcrawfor/hungar.pdf
    """

    def __init__(self, matrix):
        """
        Input a matrix and save it as a boolean matrix to designate zero locations.
        Run calculation procedure to generate results.
        """
        # Find zeros in matrix
        self._zero_locations = matrix == 0
        self._shape = matrix.shape

        # Choices starts without any choices made.
        self._choices = np.zeros(self._shape, dtype=bool)

        self._marked_rows = []
        self._marked_columns = []

        # marks rows and columns
        self.__calculate()

        # Draw lines through all unmarked rows and all marked columns.
        self._covered_rows = list(set(range(self._shape[0])) - set(self._marked_rows))
        self._covered_columns = self._marked_columns

    def get_covered_rows(self):
        """Return list of covered rows."""
        return self._covered_rows

    def get_covered_columns(self):
        """Return list of covered columns."""
        return self._covered_columns

    def __calculate(self):
        """
        Calculates minimum number of lines necessary to cover all zeros in a matrix.
        Algorithm based on: http://weber.ucsd.edu/~vcrawfor/hungar.pdf
        """
        while True:
            # Erase all marks.
            self._marked_rows = []
            self._marked_columns = []

            # Mark all rows in which no choice has been made.
            for index, row in enumerate(self._choices):
                if not row.any():
                    self._marked_rows.append(index)

            # If no marked rows then finish.
            if not self._marked_rows:
                return True

            # Mark all columns not already marked which have zeros in marked rows.
            num_marked_columns = self.__mark_new_columns_with_zeros_in_marked_rows()

            # If no new marked columns then finish.
            if num_marked_columns == 0:
                return True

            # While there is some choice in every marked column.
            while self.__choice_in_all_marked_columns():
                # Some Choice in every marked column.

                # Mark all rows not already marked which have choices in marked columns.
                num_marked_rows = self.__mark_new_rows_with_choices_in_marked_columns()

                # If no new marks then Finish.
                if num_marked_rows == 0:
                    return True

                # Mark all columns not already marked which have zeros in marked rows.
                num_marked_columns = self.__mark_new_columns_with_zeros_in_marked_rows()

                # If no new marked columns then finish.
                if num_marked_columns == 0:
                    return True

            # No choice in one or more marked columns.
            # Find a marked column that does not have a choice.
            choice_column_index = self.__find_marked_column_without_choice()

            while choice_column_index is not None:
                # Find a zero in the column indexed that does not have a row with a choice.
                choice_row_index = self.__find_row_without_choice(choice_column_index)

                # Check if an available row was found.
                new_choice_column_index = None
                if choice_row_index is None:
                    # Find a good row to accomodate swap. Find its column pair.
                    choice_row_index, new_choice_column_index = (
                        self.__find_best_choice_row_and_new_column(choice_column_index)
                    )

                    # Delete old choice.
                    self._choices[choice_row_index, new_choice_column_index] = False

                # Set zero to choice.
                self._choices[choice_row_index, choice_column_index] = True

                # Loop again if choice is added to a row with a choice already in it.
                choice_column_index = new_choice_column_index

    def __mark_new_columns_with_zeros_in_marked_rows(self):
        """Mark all columns not already marked which have zeros in marked rows."""
        num_marked_columns = 0
        for index, column in enumerate(self._zero_locations.T):
            if index not in self._marked_columns:
                if column.any():
                    (row_indices,) = np.where(column)
                    zeros_in_marked_rows = (
                        set(self._marked_rows) & set(row_indices)
                    ) != set([])
                    if zeros_in_marked_rows:
                        self._marked_columns.append(index)
                        num_marked_columns += 1
        return num_marked_columns

    def __mark_new_rows_with_choices_in_marked_columns(self):
        """Mark all rows not already marked which have choices in marked columns."""
        num_marked_rows = 0
        for index, row in enumerate(self._choices):
            if index not in self._marked_rows:
                if row.any():
                    (column_index,) = np.where(row)
                    if column_index in self._marked_columns:
                        self._marked_rows.append(index)
                        num_marked_rows += 1
        return num_marked_rows

    def __choice_in_all_marked_columns(self):
        """Return Boolean True if there is a choice in all marked columns. Returns boolean False otherwise."""
        for column_index in self._marked_columns:
            if not self._choices[:, column_index].any():
                return False
        return True

    def __find_marked_column_without_choice(self):
        """Find a marked column that does not have a choice."""
        for column_index in self._marked_columns:
            if not self._choices[:, column_index].any():
                return column_index

        raise HungarianError(
            "Could not find a column without a choice. Failed to cover matrix zeros. Algorithm has failed."
        )

    def __find_row_without_choice(self, choice_column_index):
        """Find a row without a choice in it for the column indexed. If a row does not exist then return None."""
        (row_indices,) = np.where(self._zero_locations[:, choice_column_index])
        for row_index in row_indices:
            if not self._choices[row_index].any():
                return row_index

        # All rows have choices. Return None.
        return None

    def __find_best_choice_row_and_new_column(self, choice_column_index):
        """
        Find a row index to use for the choice so that the column that needs to be changed is optimal.
        Return a random row and column if unable to find an optimal selection.
        """
        (row_indices,) = np.where(self._zero_locations[:, choice_column_index])
        for row_index in row_indices:
            (column_indices,) = np.where(self._choices[row_index])
            column_index = column_indices[0]
            if self.__find_row_without_choice(column_index) is not None:
                return row_index, column_index

        # Cannot find optimal row and column. Return a random row and column.
        from random import shuffle

        shuffle(row_indices)
        (column_index,) = np.where(self._choices[row_indices[0]])
        return row_indices[0], column_index[0]

def optimize(matrix: np.ndarray) -> np.ndarray:
    hungarian = Hungarian()
    hungarian.calculate(matrix)
    results = hungarian.get_results()
    rows = np.array([el[0] for el in results])
    cols = np.array([el[1] for el in results])
    return rows, cols



import numpy as np


class CosKMeans:
    def __init__(
        self,
        n_clusters: int = 3,
        max_iters: int = 300,
        init_count: int = 50,
        tol: float = 1e-4,
    ):
        self.n_clusters = n_clusters  # Number of clusters
        self.max_iters = max_iters  # Maximum number of iterations
        self.tol = tol  # Tolerance to declare convergence
        self.init_count = init_count

    def initialize_centers(self, X: np.ndarray) -> None:
        """
        Guessing cluster centers
        Arguments:
            - X: embeddings (num, embeddings_dim)
        Returns:
            None
        """

        def evaluate_centers(
            centers: np.ndarray, X: np.ndarray, radius: float | None = None
        ) -> float:
            """
            Evaluates cluster centers vs inputs in two steps:
            1. For each centroid computes min distance to other cluster centers
            2. Computes number of elements in half that distance
            3. Sums all elements inside these raduises and evaluates vs total number of distances
            As a result, chooses centroids with best density nearby cluster centers
            If radius is not None, uses radius instead of 1.
            Arguments:
                - centers: centroids
                - X: inputs (embeddings)
                - radius: radius around each centroid to count elements
            Returns:
                count of elements in proximity to centroids vs total number of distances
            """
            if radius is None:
                radii = self.get_radii()
            else:
                radii = [radius for _ in centers]
            distances = 1 - X @ centers.T
            elements = [
                np.sum(distances[:, i] < radii[i] / 2) for i in range(len(radii))
            ]
            return sum(elements) / np.size(distances)

        #
        options = [
            self.initialize_centers_single_pass(X) for _ in range(self.init_count)
        ]
        scores = [evaluate_centers(option, X) for option in options]
        self.cluster_centers_ = options[scores.index(max(scores))]
        self.cluster_centers_ = self.norm(self.cluster_centers_)

    def initialize_centers_single_pass(self, X: np.ndarray) -> np.ndarray:
        """
        Single K-means cluster centers run
        Arguments:
            - inputs (embeddings)
        Returns:
            centers of clusters
        """
        # Step 1: Initialize the first center randomly
        centers = np.expand_dims(X[np.random.choice(X.shape[0])], 0)

        # Step 2: Initialize the remaining centers using k-means++ method
        for _ in range(1, self.n_clusters):
            # Compute the distance of each point to the nearest center
            # distances = np.array(
            #     [min(np.linalg.norm(x - c) for c in centers) for x in X]
            # )
            min_dist = np.min(1 - centers @ X.T, axis=0)
            # Probability distribution for selecting the next center
            probs = min_dist**2 / np.sum(min_dist**2)

            # Choose the next center based on the probability distribution
            next_center = X[np.random.choice(X.shape[0], p=probs)]
            # centers.append(next_center)
            centers = np.vstack((centers, next_center))

        self.cluster_centers_ = centers
        return centers

    def norm(self, X: np.ndarray, eps=1e-8) -> np.ndarray:
        """
        Returns normalized vectors
        Arguments:
            - X: inputs
        Returns:
            numpy array of normalized inputs
        """
        norm = np.expand_dims(np.linalg.norm(X, axis=1), axis=1)
        norm[norm == 0] = (np.ones_like(norm) * eps)[norm == 0]
        return X / norm

    def fit(self, X: np.ndarray) -> None:
        """
        Fit pass without prediction
        Arguments:
            - X: inputs (embeddings)
        Returns:
            None
        """
        self.initialize_centers(X)

        for _ in range(self.max_iters):
            distances = 1 - X @ self.cluster_centers_.T
            self.labels_ = np.argmin(distances, axis=1)
            new_centers = np.array(
                [X[self.labels_ == j].mean(axis=0) for j in range(self.n_clusters)]
            )
            new_centers = self.norm(new_centers)
            metric = np.mean(1 - np.max(new_centers @ self.cluster_centers_.T, axis=1))
            if metric < self.tol:
                break
            self.cluster_centers_ = new_centers

    def predict(self, X: np.ndarray) -> np.ndarray:
        """
        Prediction pass without fitting
        Arguments:
            - X: inputs (embeddings)
        Returns:
            numpy array of labels
        """
        distances = 1 - X @ self.cluster_centers_.T
        return np.argmin(distances, axis=1)

    def predict_with_confidence(self, X: np.ndarray) -> np.ndarray:
        """
        Prediction pass without fitting, returns numpy array of weighted
        predictions per each sample per each cluster
        Arguments:
            - X: inputs (embeddings)
        Returns:
            numpy array of weighted predictions (num_samples x num_clusters)
        """
        distances = 1 - X @ self.cluster_centers_.T
        # preds = 1 / (distances / np.min(distances, axis=1, keepdims=1))
        # weights = 1 - (distances / np.sum(distances, axis=1, keepdims=1))
        # return preds * weights
        return (1 / distances) / np.sum(1 / distances, axis=1, keepdims=True)

    def fit_predict(self, X):
        """
        Fit + predict
        """
        self.fit(X)
        return self.predict(X)

    def fit_predict_with_confidence(self, X):
        """
        Fit + predict
        """
        self.fit(X)
        return self.predict_with_confidence(X)

    def get_radii(self):
        """
        Computes cluster radii
        Arguments:
            None
        Returns:
            numpy array of radii around each centroid
        """
        assert len(self.cluster_centers_) > 0, "No speakers to compute radii"
        if len(self.cluster_centers_) == 1:
            return [np.array(2.0)]
        ccdists = 1 - self.cluster_centers_ @ self.cluster_centers_.T
        ccc = [np.concatenate((row[:i], row[i + 1 :])) for i, row in enumerate(ccdists)]
        radii = [np.min(el) for el in ccc]
        return radii



import numpy as np
from .segment import Segments, Segment
from .hungarian import optimize
import warnings


def total_length(segments: Segments):
    return sum([segment.end - segment.start for segment in segments])


def intersection_length(a: Segment, b: Segment):
    """Compute the intersection length of two Segments."""
    max_start = max(a.start, b.start)
    min_end = min(a.end, b.end)
    return max(0.0, min_end - max_start)


def invert_segments(segments: Segments, maxtime: float | None = None):
    negative = Segments()
    start = 0.0
    for s in segments:
        candidate = Segment(start=start, end=s.start, speaker="")
        if candidate.len() > 0:
            negative.append(candidate)
        start = s.end
    if maxtime is not None and negative.max() < maxtime:
        negative.append(Segment(start=negative.max(), end=maxtime))
    return negative


def false_alarm(segments: Segments, ground_truth: Segments):
    ngt = invert_segments(ground_truth)
    fa = 0.0

    for s in segments:
        for gt in ngt:
            fa += intersection_length(s, gt)
    return fa


def missed_detection(segments: Segments, ground_truth: Segments):
    ###   ground truth: ........xxxxxxxxxx.......xxxxxxxx........
    ###    predictions: .....xxxxxxx.......xxxxxxxxx.............
    ###         missed:             ++++++          +++++

    ns = invert_segments(segments, ground_truth.max())
    md = 0.0

    for gt in ground_truth:
        for s in ns:
            md += intersection_length(s, gt)
    return md


def overlap(a: Segments, b: Segments) -> float:
    ov = 0.0
    for sa in a:
        for sb in b:
            ov += intersection_length(sa, sb)
    return ov


def confusion(
    segments: Segments, ground_truth: Segments, rows: np.ndarray, cols: np.ndarray
):
    cnf = 0.0
    rows = list(rows)
    cols = list(cols)
    gtspeakers = ground_truth.speakers()
    for i, speaker in enumerate(segments.speakers()):
        sps = [el for el in segments if el.speaker == speaker]
        gtspeaker_id = cols[rows.index(i)]
        gtps = [el for el in ground_truth if el.speaker != gtspeakers[gtspeaker_id]]
        for s in sps:
            for gt in gtps:
                if (ilen := intersection_length(s, gt)) > 0:
                    cnf += ilen

    return cnf


def speaker_index(segments: Segments):
    """Build the index for the speakers.

    Args:
        hyp: a list of tuples, where each tuple is (speaker, start, end)
            of type (string, float, float)

    Returns:
        a dict from speaker to integer
    """
    speaker_set = sorted({seg.speaker for seg in segments})
    index = {speaker: i for i, speaker in enumerate(speaker_set)}
    return index


def compute_merged_total_length(ref: Segments, hyp: Segments):
    """Compute the total length of the union of reference and hypothesis.

    Arguments:
        - ref: a list of tuples for the ground truth, where each tuple is
            (speaker, start, end) of type (string, float, float)
        - hyp: a list of tuples for the diarization result hypothesis, same type
            as `ref`

    Returns:
        a float number for the union total length
    """
    # Remove speaker label and merge.
    merged = [(seg.start, seg.end) for seg in (ref + hyp)]
    # Sort by start.
    merged = sorted(merged, key=lambda el: el[0])
    i = len(merged) - 2
    while i >= 0:
        if merged[i][1] >= merged[i + 1][0]:
            max_end = max(merged[i][1], merged[i + 1][1])
            merged[i] = (merged[i][0], max_end)
            del merged[i + 1]
            if i == len(merged) - 1:
                i -= 1
        else:
            i -= 1
    total_length = sum([el[1] - el[0] for el in merged])
    return total_length


def cost_matrix(reference: Segments, hypothesis: Segments):
    """Build the cost matrix.

    Arguments:
        - ref: a list of tuples for the ground truth, where each tuple is
            (speaker, start, end) of type (string, float, float)
        - hyp: a list of tuples for the diarization result hypothesis, same type
            as `ref`

    Returns:
        a 2-dim numpy array, whose element (i, j) is the overlap between
            `i`th reference speaker and `j`th hypothesis speaker
    """
    assert len(reference.speakers()) == len(
        hypothesis.speakers()
    ), "Number of speakers between reference and hypothesis do not match"
    reference_index = speaker_index(reference)
    hypothesis_index = speaker_index(hypothesis)

    cmatrix = np.zeros((len(reference_index), len(hypothesis_index)))
    for reference_el in reference:
        for hypothesis_el in hypothesis:
            i = reference_index[reference_el.speaker]
            j = hypothesis_index[hypothesis_el.speaker]
            cmatrix[i, j] += intersection_length(reference_el, hypothesis_el)
    return cmatrix


def DER(reference: Segments, hypothesis: Segments) -> float:
    def check_overlap(reference):
        rs = reference.speakers()
        others = rs.copy()
        for speaker in rs:
            others.remove(speaker)
            for other in others:
                for sa in reference.filter(speaker):
                    for sb in reference.filter(other):
                        if intersection_length(sa, sb) > 0:
                            return True
        return False

    if check_overlap(reference=reference):
        warnings.warn(
            "Current DER implementation does not support multispeaker with overlap. Try using other packages."
        )
    ref_duration = total_length(reference)
    cmatrix = cost_matrix(reference, hypothesis)
    rows, cols = optimize(-cmatrix)

    fa = false_alarm(segments=hypothesis, ground_truth=reference)
    correct = cmatrix[rows, cols].sum()
    total = ref_duration
    md = missed_detection(segments=hypothesis, ground_truth=reference)
    cnf = confusion(segments=hypothesis, ground_truth=reference, rows=rows, cols=cols)

    # union_total_length = compute_merged_total_length(reference, hypothesis)
    # der = (union_total_length - correct) / ref_duration
    der = (fa + md + cnf) / ref_duration
    return {
        "false alarm": fa,
        "correct": correct,
        "total": total,
        "missed detection": md,
        "confusion": cnf,
        "diarization error rate": der,
    }



import numpy as np


def gaussian_kernel(size, sigma):
    """Generates a 1D Gaussian kernel."""
    kernel = np.fromfunction(
        lambda x: (1 / (np.sqrt(2 * np.pi) * sigma))
        * np.exp((-((x - (size - 1) / 2) ** 2)) / (2 * sigma**2)),
        (size,),
    )
    return kernel / np.sum(kernel)


def convolve1d(input_array, kernel):
    """Convolves a 1D input array with a 1D kernel."""
    input_size = len(input_array)
    kernel_size = len(kernel)
    pad_width = kernel_size // 2

    # Pad the input array with reflect mode
    padded_input = np.pad(input_array, pad_width, mode="reflect")

    # Initialize the output array
    output_array = np.zeros(input_size)

    # Perform convolution
    for i in range(input_size):
        output_array[i] = np.sum(padded_input[i : i + kernel_size] * kernel)

    return output_array


def gaussian_filter1d(input_array, sigma):
    """Applies a 1D Gaussian filter to a 1D input array."""
    # Define the size of the kernel (6 * sigma is a common choice)
    kernel_size = int(6 * sigma + 1)
    if kernel_size % 2 == 0:
        kernel_size += 1  # Ensure kernel size is odd

    # Generate Gaussian kernel
    kernel = gaussian_kernel(kernel_size, sigma)

    # Apply convolution
    return convolve1d(input_array, kernel)


if __name__ == "__main__":
    pass



from .main import modelinfo, Jadia
from .audio import SampleProcessor
from .models import Model, model_lite
from .segment import Segment, Segments
from .coskmeans import CosKMeans
from .metrics import cost_matrix, DER



from glob import glob
import os
from jadia import Jadia, Segments
import jadia_plot as plot
import torch
import warnings
from datetime import datetime
import pickle
from pyannote.metrics.diarization import DiarizationErrorRate
from pyannote.audio import Pipeline
from pyannote.database.util import load_lab

# Load Jadia
diarizator = Jadia(
    device=torch.device("cuda:0"),
    model="lite_v2",
    batch_size=64,
)

# load pyannote DER function
pyannoteDER = DiarizationErrorRate()

# dirs
audio_dir = "tests/assets"
ext = ".wav"
ref_rttm_dir = "tests/assets"
target_dir = "results/"

try:
    os.mkdir(target_dir)
    os.mkdir(target_dir + "/images")
except:
    pass


# config
COLLAPSE = True
COLLAPSE_TOLERANCE = 0.7
USE_VAD = True
ADJUST_TOLERANCE = 0.7
MIN_INTERVAL = 0.2
SMOOTH = False
FAST_FIT = False
SPECTROGRAM_STRIDE = 5
UPDATE_STRIDE = 7
CLUSTERING_METHOD = "CosKMeans"
WITH_CONFIDENCE = True

# Use pyannote?
RUN_PYANNOTE = True

if RUN_PYANNOTE is True:
    pipeline = Pipeline.from_pretrained(
        "pyannote/speaker-diarization-3.1",
        use_auth_token="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",
    ).to(torch.device("cuda"))


def test_eval():

    # load audio files
    files = sorted(glob(audio_dir + f"/*{ext}"))
    stats = {}
    for audio_fpath in files:
        print(f"Processing {os.path.basename(audio_fpath)}")
        print(f"  Collapse: {COLLAPSE}, VAD: {USE_VAD}, Tolerance: {ADJUST_TOLERANCE}")
        print(f"  Min interval: {MIN_INTERVAL}, Smooth: {SMOOTH}, Fast fit: {FAST_FIT}")
        name = ".".join(os.path.basename(audio_fpath).split(".")[:-1])

        # if os.path.isfile(f"{images_dir}/{name}_{CLUSTERING_METHOD}.png"):
        #    continue

        ref_rttm_fpath = os.path.join(ref_rttm_dir, name) + ".rttm"
        try:
            reference = Segments().from_rttm_load(ref_rttm_fpath)
        except:
            continue
        num_voices = len(reference.speakers())
        diarizator.setup(num_voices=num_voices, with_confidence=True)

        # Compute predictions [Jadia]
        jst = datetime.now()
        diarizator.load_audio(filename=audio_fpath)
        predictions = diarizator.predict(fast_fit=True)
        segments = diarizator.predictions_to_segments(predictions)
        jet = datetime.now()

        if RUN_PYANNOTE:
            # Compute predictions [Pyannote]
            pst = datetime.now()
            pydia = pipeline(audio_fpath)
            pet = datetime.now()
            with open("pypred.rttm", "w") as f:
                pydia.write_rttm(f)

        segments.to_rttm_save(
            fpath=os.path.join(target_dir, name) + ".rttm", overwrite=True
        )
        reference.to_lab_save("ref.lab")
        segments.to_lab_save("hyp.lab")

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            ref = load_lab("ref.lab")
            hyp = load_lab("hyp.lab")
            der = pyannoteDER(ref, hyp, detailed=True)
            os.remove("ref.lab")
            os.remove("hyp.lab")

        if RUN_PYANNOTE:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore")
                der2 = pyannoteDER(ref, pydia, detailed=True)
                # der2_ = DER(ref, pydia)

        # plot combined chart
        if RUN_PYANNOTE:
            segments2 = Segments()
            segments2.from_rttm_load("pypred.rttm")
            os.remove("pypred.rttm")
        else:
            segments2 = None

        plot.plot_predictions(
            predictions=predictions,
            segments=segments,
            segments2=segments2,
            filename=f"{target_dir}/images/{name}_{CLUSTERING_METHOD}.png",
            ground_truth=reference,
        )

        plot.plot_segments(
            pred=segments,
            ground_truth=reference,
            filename=f"{target_dir}/images/{name}_{CLUSTERING_METHOD}_iv.png",
        )

        print("-----------[RESULTS: Jadia]------------")
        print(f"       Confusion: {der['confusion']:.4f}")
        print(f"     False alarm: {der['false alarm']:.4f}")
        print(f"Missed detection: {der['missed detection']:.4f}")
        print(f"             DER: {der['diarization error rate']:.4f}")
        print(f"            Time: {(jet-jst).total_seconds()}")

        if RUN_PYANNOTE:
            print("-----------[RESULTS: Pyannote]------------")
            print(f"       Confusion: {der2['confusion']:.4f}")
            print(f"     False alarm: {der2['false alarm']:.4f}")
            print(f"Missed detection: {der2['missed detection']:.4f}")
            print(f"             DER: {der2['diarization error rate']:.4f}")
            print(f"            Time: {(pet-pst).total_seconds()}")

        stats[name] = {
            "confusion": {
                "Jadia": der["confusion"],
            },
            "false_alarm": {
                "Jadia": der["false alarm"],
            },
            "missed_detection": {
                "Jadia": der["missed detection"],
            },
            "der": {
                "Jadia": der["diarization error rate"],
            },
            "time": {
                "Jadia": (jet - jst).total_seconds() * 1000,
            },
        }
        if RUN_PYANNOTE:
            stats[name]["confusion"]["pyannote"] = der2["confusion"]
            stats[name]["false_alarm"]["pyannote"] = der2["false alarm"]
            stats[name]["missed_detection"]["pyannote"] = der2["missed detection"]
            stats[name]["der"]["pyannote"] = der2["diarization error rate"]
            stats[name]["time"]["pyannote"] = (pet - pst).total_seconds() * 1000

    with open(os.path.join(target_dir, "stats.pkl"), "wb") as fp:
        pickle.dump(stats, fp)


if __name__ == "__main__":
    test_eval()



import numpy as np
import torch
from torch.utils.data import Dataset, DataLoader
from .coskmeans import CosKMeans
from . import models as _m
from .audio import SampleProcessor
from .segment import Segment, Segments
from .gaussian import gaussian_filter1d
import warnings
import os

# from sklearn.cluster import KMeans, AgglomerativeClustering, DBSCAN


modelinfo = {
    "lite": {
        "name": "model_lite",
        "checkpoint": "assets/lite.pt",
    },
    "lite_v2": {
        "name": "model_lite",
        "checkpoint": "assets/lite_v2.pt",
    },
}


class Jadia:
    def __init__(
        self,
        device: torch.device = torch.device("cpu"),
        model: str = "lite",
        clustering_method: str = "kmeans",
        batch_size: int = 16,
    ) -> None:
        self.device = device
        self.batch_size = batch_size
        self.model = Model(device=device, model=model)
        self.clustering_method = clustering_method

    def setup(
        self,
        num_voices: int,
        use_vad: bool = True,
        collapse: bool = True,
        collapse_tolerance: float = 0.5,
        adjust_tolerance: float = 0.5,
        min_interval: float = 0.1,
        spectrogram_stride: int = 10,
        spectrogram_length: float = 60,
        spectrogram_strategy: str = "pad",
        update_stride: int = 1,
        clustering_method: str = "coskmeans",
        with_confidence: bool = False,
        **kwargs,
    ):
        """
        Sets up class instance for the next audio to process

        Arguments:
            - num_voices: number of speakers
            - use_vad: use voice activity detection
            - collapse: ignore non-spoken intervals in between when merging same-speaker segments
            - collapse_tolerance: max size of non-spoken intervals to collapse, in seconds
            - adjust_tolerance: edge-wise adjust predictions, in seconds
            - min_interval: min spoken interval, in seconds
            - spectrogram_stride: sliding window stride over spectrogram
            - spectrogram_strategy: pad or tile, how to reconstruct sliding window if spectrogram is too small
            - update_stride: populate predictions by average within stride interval
            - clustering_method: currently only cosine KMeans (coskmeans)
            - kwargs: audio args, see SampleProcessor class init params
        Returns:
            None
        """
        self.num_voices = num_voices
        self.collapse = collapse
        self.collapse_tolerance = collapse_tolerance
        self.tolerance = adjust_tolerance
        self.use_vad = use_vad
        self.min_interval = min_interval
        self.audio_args = kwargs
        self.audio_args["stride"] = spectrogram_stride
        self.audio_args["slice"] = spectrogram_length
        self.audio_args["strategy"] = spectrogram_strategy
        self.update_stride = update_stride
        self.update_fn = (
            self.update_predictions_with_confidence
            if with_confidence is True
            else self.update_predictions
        )
        self.spectrogram_strategy = spectrogram_strategy
        self.clusterer = Clusterer(num_voices=num_voices, method=clustering_method)

    def load_audio(self, filename: str) -> SampleProcessor:
        """
        Loads audio, detects silent|non-spoken parts and creates spectrogram
        Arguments:
            - filename: path to audio file
        Returns:
            - SampleProcessor instance
        """
        self.filename = filename
        self.sample_processor = SampleProcessor(
            filename, use_vad=self.use_vad, **self.audio_args
        )
        return self.sample_processor

    def predict(
        self,
        num_voices: int | None = None,
        smooth: bool = True,
        fast_fit: bool = False,
    ) -> np.ndarray:
        """
        Converts loaded audio to dict of predictions,
        where prediction keys are spectrogram-wise positions
        Arguments:
            - num_voices: number of speakers
            - smooth: use gaussian filter on predictions, may increase accuracy of speaker change around silences
            - fast_fit: compute clusters based on first part of audio, don't collect all embeddings. Good for long dialogues. You may need to increase length of slice (60s by default).
        Returns:
            predictions, numpy array (spectrogram length x num_voices)
        """
        if num_voices is None:
            num_voices = self.num_voices

        # create predictions array with shape (combined_spectrogram_len x num_voices)
        self.predictions = np.zeros(
            (self.sample_processor.get_fullspectrogram_size(), num_voices)
        )

        embeddings = {}
        # for index in count spectrograms
        for index in range(self.sample_processor.count_spectrograms()):
            # create dataset made of sliding windows across that spectrogram
            dataset = DTST(
                data=self.sample_processor.get_sliding_windows(index),
                device=self.device,
            )
            loader = DataLoader(dataset, batch_size=self.batch_size)
            # compute and populate embeddings per sliding window
            embs = []
            with torch.no_grad():
                for inputs in loader:
                    be = self.model(inputs)
                    embs.append(be.cpu().numpy())
            embs = np.vstack(embs)
            # normalize embeddings (project onto unit hypersphere)
            embeddings[index] = embs / np.expand_dims(np.linalg.norm(embs, axis=1), 1)
            # if we compute centroids using first spectrogram
            if fast_fit is True:
                if index == 0:
                    assert (
                        len(embeddings[index]) >= num_voices
                    ), "Unable to process audio, number of embeddings generated less than number of voices [too little embeddings to compute clusters]. You may try setting fast_fit to False, or use vad=False for setup() or process()"
                self.update_fn(
                    indices=[index],
                    embeddings=[embeddings[index]],
                    fit=True if index == 0 else False,
                    stride=self.update_stride,
                )
                embeddings = {}
        # if we compute centroids using combined spectrogram
        if fast_fit is False:
            assert (
                sum([len(embs) for embs in embeddings.values()]) >= num_voices
            ), "Unable to process audio, number of embeddings generated less than number of voices [too little embeddings to compute clusters]. You may try using vad=False for setup() or process()"
            self.update_fn(
                indices=list(embeddings.keys()),
                embeddings=list(embeddings.values()),
                fit=True,
                stride=self.update_stride,
            )
        if smooth is True:
            self.smooth_predictions()
        return self.predictions

    def update_predictions(
        self,
        indices: list[int],
        embeddings: list[np.ndarray],
        fit: bool = False,
        stride: int = 1,
    ):
        """
        Updates predictions based on embeddings
        Arguments:
            - indices: list of spectrograms to update predictions against
            - embeddings: according embeddings
            - fit: fit+predict or just predict without fitting
            - stride: average using stride for predictions
        Returns:
            None
        """
        if fit is True:
            labels = self.clusterer.fit_predict(np.vstack(embeddings))
            if self.num_voices > len(np.unique(labels)):
                warnings.warn(
                    "Number of clusters is less than number of speakers. Try setting fast_fit to false if you're using it."
                )
            lengths = [emb.shape[0] for emb in embeddings]
            starts = np.cumsum([0] + lengths[:-1])
            predictions = np.split(
                labels, [start + length for start, length in zip(starts, lengths)]
            )
        else:
            predictions = [self.clusterer.predict(np.vstack(embeddings))]

        for seq_index, preds_per_seq_index in zip(indices, predictions):
            windows_nums, matching_weights = (
                self.sample_processor.sliding_window_indices(seq_index)
            )
            for global_interval, local_interval in self.sample_processor.iter_intervals(
                seq_index=seq_index
            ):
                embedding_indices = windows_nums[
                    local_interval[0] : local_interval[1] + 1
                ]
                weights = matching_weights[local_interval[0] : local_interval[1] + 1]
                for i in range(0, local_interval[1] - local_interval[0] + 1, stride):
                    t = i + global_interval[0]
                    if len(embedding_indices[i]) == 0:
                        # we can have no windows for some positions due to large stride
                        continue
                    preds = preds_per_seq_index[embedding_indices[i]]
                    sw = sum(weights[i])
                    for s in np.unique(preds):
                        self.predictions[t : t + stride, s] = (
                            np.sum(weights[i][preds == s]) / sw
                        )

    def update_predictions_with_confidence(
        self,
        indices: list[int],
        embeddings: list[np.ndarray],
        fit: bool = False,
        stride: int = 1,
    ):
        """
        Updates predictions based on embeddings
        Arguments:
            - indices: list of spectrograms to update predictions against
            - embeddings: according embeddings
            - fit: fit+predict or just predict without fitting
            - stride: average using stride for predictions
        Returns:
            None
        """
        if fit is True:
            # compute centroids and return predictions with confidence,
            # - combined across all spectrograms if multiple
            clusterer_estimates = self.clusterer.fit_predict_with_confidence(
                np.vstack(embeddings)
            )
            if self.num_voices > clusterer_estimates.shape[1]:
                warnings.warn(
                    "Number of clusters is less than number of speakers. Try setting fast_fit to false if you're using it."
                )
            # get number of embeddings per spectrogram
            lengths = [emb.shape[0] for emb in embeddings]
            starts = np.cumsum([0] + lengths[:-1])
            # split estimates with confidence back into groups per spectrogram
            estimates = np.split(
                clusterer_estimates,
                [start + length for start, length in zip(starts, lengths)],
            )
        else:
            estimates = [self.clusterer.predict_with_confidence(np.vstack(embeddings))]

        # for each spectgrogram get all according estimates
        for spg_index, estimates_per_spg_index in zip(indices, estimates):
            # get ids of sliding windows and matching weights
            # per each position within spectrogram
            windows_nums, matching_weights = (
                self.sample_processor.sliding_window_indices(spg_index)
            )
            # for each interval withing spectrogram we get position of such interval
            # but on a global audio sample (in spectrogram steps, not in pure audio samples)
            for global_interval, local_interval in self.sample_processor.iter_intervals(
                seq_index=spg_index
            ):
                # get indices of embeddings/estimates per position + according weights
                embedding_indices = windows_nums[
                    local_interval[0] : local_interval[1] + 1
                ]
                weights = matching_weights[local_interval[0] : local_interval[1] + 1]
                # for each strided position within local interval
                for i in range(0, local_interval[1] - local_interval[0] + 1, stride):
                    # compute according position within global interval
                    t = i + global_interval[0]
                    # we can have no sliding windows for some positions due to large stride
                    if len(embedding_indices[i]) == 0:
                        continue
                    # get model estimates for each sliding window per position
                    ests = estimates_per_spg_index[embedding_indices[i]]
                    # apply weights assigned to each sliding widow to estimates
                    mm = np.expand_dims(weights[i], axis=1) * ests
                    self.predictions[t : t + stride, :] = np.expand_dims(
                        np.sum(mm, axis=0) / np.sum(mm), axis=0
                    )

    def process(
        self,
        filename: str,
        num_voices: int,
        use_vad: bool = True,
        fast_fit: bool = False,
        **kwargs,
    ) -> Segments:
        """
        Simplified combined function to process audio from loading to predictions
        Arguments (see Sample class constructor for default values):
            - filename:
            - num_voices: number of speakers
            - use_vad: use voice activity detection
            - fast_fit: fit using part of record (60 seconds by default)
            + kwargs for SampleProcessor
        Returns:
            instance of Segments
        """
        self.setup(num_voices=num_voices, use_vad=use_vad, **kwargs)
        self.sample_processor = self.load_audio(filename=filename)
        predictions = self.predict(num_voices=num_voices, fast_fit=fast_fit)
        segments = self.predictions_to_segments(predictions)
        return segments

    def follow(self, filename: str) -> list:
        """
        Function similar to process, but it uses previous clusterer and audio settings.
        Arguments:
            - filename: str
        Returns:
            instance of Segments
        """
        self.sample_processor = self.load_audio(filename=filename)
        predictions = self.predict(num_voices=self.num_voices, smooth=True)
        segments = self.predictions_to_segments(predictions, tolerance=self.tolerance)
        return segments

    def smooth_predictions(self):
        """
        Smoothes predictions, helps with incorrect speaker change predictions near silent parts
        Arguments:
            none
        Returns:
            none
        """
        # Parameters for the Gaussian kernel
        for s in range(len(self.predictions[0])):
            arr = gaussian_filter1d(
                self.predictions[:, s], sigma=15
            )  # apply_convolution([el[s] for el in predictions.values()], kernel)
            arr[~self.sample_processor.mask] = 0
            self.predictions[:, s] = arr

    def predictions_to_segments(
        self,
        predictions: np.ndarray,
    ):
        """
        Converts dictionary of predictions to Segments instance
        Arguments:
            - predictions: dict of predictions from previous computations
        Returns:
            instance of Segments
        """
        assert hasattr(
            self, "collapse"
        ), "Need to run setup() before calling this method"
        best = np.argmax(predictions, axis=1)
        best[np.max(predictions, axis=1) == 0] = -1
        segments = Segments()
        change_indices = np.where(np.diff(best) != 0)[0] + 1
        change_indices = np.concatenate(([0], change_indices, [len(best)]))
        for start, end in zip(change_indices[:-1], change_indices[1:]):
            segments.append(
                Segment(
                    start=self.sample_processor.step_to_time(start),
                    end=self.sample_processor.step_to_time(end - 1),
                    speaker=(best[start] if best[start] > -1 else None),
                )
            )

        if self.tolerance > 0:
            # adjust segments:
            # if interval close to silence (but not cigar), followed by different speaker,
            # shift edge towards silence
            edges = np.array(
                [
                    self.sample_processor.step_to_time(el)
                    for iv in self.sample_processor.get_intervals()
                    for el in iv
                ]
            )
            segments.adjust(edges=edges, tolerance=self.tolerance)
        # drop too short segments
        segments.drop_short(self.min_interval)
        # merge segments
        segments.merge(collapse=self.collapse, tolerance=self.collapse_tolerance)
        # remove silent segments
        segments.drop_silent()

        return segments

    def order_speakers(self):
        """
        Modifies speakers' order, makes them appear in same order as regular enumeration
        """
        mv = np.argmax(self.predictions, axis=1)[
            np.argwhere(np.max(self.predictions, axis=1) > 0)
        ]
        _, ixs = np.unique(mv, return_index=True)
        self.predictions = self.predictions[:, np.argsort(ixs)]


class Model:
    def __init__(
        self, device: torch.device = torch.device("cpu"), model: str = "lite"
    ) -> None:
        self.device = device
        if model not in modelinfo.keys():
            raise Exception("Unkown model name")
        model_call = getattr(_m, modelinfo[model]["name"])
        self.model = model_call().to(device)
        self.model.eval()
        cp = os.path.dirname(__file__)
        with open(os.path.join(cp, modelinfo[model]["checkpoint"]), "rb") as f:
            checkpoint = torch.load(f, weights_only=True)
        self.model.load_state_dict(checkpoint)

    def __call__(self, inputs: torch.Tensor) -> torch.Tensor:
        return self.model(inputs)


class DTST(Dataset):
    def __init__(
        self,
        data: torch.Tensor | np.ndarray,
        dtype: torch.dtype = torch.float32,
        device: str | torch.device = "cpu",
    ) -> None:
        super().__init__()
        if isinstance(data, np.ndarray):
            self.data = torch.from_numpy(data).to(dtype=dtype)
        else:
            self.data = data.to(dtype=dtype)
        self.dtype = dtype
        self.device = device

    def __len__(self) -> int:
        return self.data.shape[0]

    def __getitem__(self, index) -> torch.Tensor:
        return self.data[index].to(device=self.device)


class Clusterer:
    def __init__(
        self, num_voices: int | None = None, method: str = "coskmeans"
    ) -> None:
        self.method = method.lower()
        """
        if self.method == "kmeans":
            assert (
                num_voices is not None
            ), "With K-means you need to specify number of speakers"
            self.engine = KMeans(n_clusters=num_voices)
        """
        if self.method == "coskmeans":
            assert (
                num_voices is not None
            ), "With CosKmeans you need to specify number of speakers"
            self.engine = CosKMeans(n_clusters=num_voices)
            """
            elif self.method == "agg":
                assert (
                    num_voices is not None
                ), "With Agglomerative clustering you need to specify number of speakers"
                self.engine = AgglomerativeClustering(
                    n_clusters=num_voices, metric="cosine", linkage="average"
                )
            elif self.method == "dbscan":
                assert (
                    num_voices is not None
                ), "With Agglomerative clustering you need to specify number of speakers"
                self.engine = DBSCAN(eps=1e-2, metric="cosine")
            """
        else:
            raise Exception(
                "Unknown clustering method, only kmeans, agg or dbscan are supported"
            )

    def fit_predict(self, embeddings: np.ndarray) -> np.ndarray:
        labels = self.engine.fit_predict(embeddings)
        # if self.method == "kmeans" or self.method == "coskmeans":
        self.centroids = self.engine.cluster_centers_
        """
        elif self.method == "agg" or self.method == "dbscan":
            self.__refdata = embeddings
            self.__reflabels = labels
        """
        return labels

    def fit_predict_with_confidence(self, embeddings: np.ndarray) -> np.ndarray:
        assert self.method == "coskmeans", "Only CosKMeans supported with this method"
        return self.engine.fit_predict_with_confidence(embeddings)

    def predict(self, embeddings: np.ndarray) -> np.ndarray:
        return self.engine.predict(embeddings)

    def predict_with_confidence(self, embeddings: np.ndarray) -> np.ndarray:
        return self.engine.predict_with_confidence(embeddings)



import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from typing import Optional
from torch import Tensor


class DirectionalConv2d(nn.Module):
    def __init__(
        self,
        shape: tuple,
        in_channels: int,
        out_channels: int,
        kernel: int | tuple,
        stride: int | tuple = 1,
        padding: int | tuple = 0,
        groups: int = 1,
        bias: bool = True,
        extras: list | None = None,
    ) -> None:
        super().__init__()
        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=kernel,
            stride=stride,
            padding=padding,
            groups=groups,
            bias=bias,
        )
        self.bn = nn.BatchNorm2d(out_channels)
        self.bn2 = nn.BatchNorm2d(out_channels)
        self.activation = nn.ReLU()
        self.extras = nn.Sequential(*extras) if extras is not None else None
        self.alignment = DirectionalAlignment(out_channels, shape)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)
        x = self.bn(x)
        x = self.activation(x)
        if self.extras is not None:
            x = self.extras(x)
        y = x.clone()
        y = self.alignment(y)
        x = x + y
        x = self.bn2(x)
        return x


class DirectionalAlignment(nn.Module):
    def __init__(self, planes: int, shape: tuple) -> None:
        super().__init__()
        self.weights = nn.Parameter(torch.Tensor(planes, shape[0], shape[0]))
        nn.init.kaiming_normal_(self.weights)
        self.bn = nn.BatchNorm2d(planes)
        self.dim = torch.sqrt(torch.tensor(shape[-1]))
        self.softmax = nn.Softmax(dim=1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x @ x.transpose(-1, -2)
        x = x / self.dim
        x = x * self.weights
        x = torch.sum(x, dim=-1, keepdim=True)
        return x

    def extra_repr(self) -> str:
        return f"DirectionalAlignment: planes={self.weights.shape[0]}, shape={self.weights.shape[1:]}"


def sinusoids(length, channels, max_timescale=10000):
    """Returns sinusoids for positional embedding"""
    assert channels % 2 == 0
    log_timescale_increment = np.log(max_timescale) / (channels // 2 - 1)
    inv_timescales = torch.exp(-log_timescale_increment * torch.arange(channels // 2))
    scaled_time = torch.arange(length)[:, np.newaxis] * inv_timescales[np.newaxis, :]
    return torch.cat([torch.sin(scaled_time), torch.cos(scaled_time)], dim=1)


class MultiHeadAttention(nn.Module):
    def __init__(self, n_state: int, n_head: int):
        super().__init__()
        self.n_head = n_head
        self.query = nn.Linear(n_state, n_state)
        self.key = nn.Linear(n_state, n_state, bias=False)
        self.value = nn.Linear(n_state, n_state)
        self.out = nn.Linear(n_state, n_state)

    def forward(
        self,
        x: Tensor,
        xa: Optional[Tensor] = None,
        mask: Optional[Tensor] = None,
        kv_cache: Optional[dict] = None,
    ):
        q = self.query(x)

        if kv_cache is None or xa is None or self.key not in kv_cache:
            # hooks, if installed (i.e. kv_cache is not None), will prepend the cached kv tensors;
            # otherwise, perform key/value projections for self- or cross-attention as usual.
            k = self.key(x if xa is None else xa)
            v = self.value(x if xa is None else xa)
        else:
            # for cross-attention, calculate keys and values once and reuse in subsequent calls.
            k = kv_cache[self.key]
            v = kv_cache[self.value]

        wv, qk = self.qkv_attention(q, k, v, mask)
        return self.out(wv), qk

    def qkv_attention(
        self,
        q: Tensor,
        k: Tensor,
        v: Tensor,
        mask: Optional[Tensor] = None,
    ):
        n_batch, n_ctx, n_state = q.shape
        scale = (n_state // self.n_head) ** -0.25
        q = q.view(*q.shape[:2], self.n_head, -1).permute(0, 2, 1, 3) * scale
        k = k.view(*k.shape[:2], self.n_head, -1).permute(0, 2, 3, 1) * scale
        v = v.view(*v.shape[:2], self.n_head, -1).permute(0, 2, 1, 3)

        qk = q @ k
        if mask is not None:
            qk = qk + mask[:n_ctx, :n_ctx]
        qk = qk.float()

        w = F.softmax(qk, dim=-1).to(q.dtype)
        return (w @ v).permute(0, 2, 1, 3).flatten(start_dim=2), qk.detach()


class ResidualAttentionBlock(nn.Module):
    def __init__(self, n_state: int, n_head: int, cross_attention: bool = False):
        super().__init__()

        self.attn = MultiHeadAttention(n_state, n_head)
        self.attn_ln = nn.LayerNorm(n_state)

        self.cross_attn = (
            MultiHeadAttention(n_state, n_head) if cross_attention else None
        )
        self.cross_attn_ln = nn.LayerNorm(n_state) if cross_attention else None

        n_mlp = n_state * 4
        self.mlp = nn.Sequential(
            nn.Linear(n_state, n_mlp), nn.GELU(), nn.Linear(n_mlp, n_state)
        )
        self.mlp_ln = nn.LayerNorm(n_state)

    def forward(
        self,
        x: Tensor,
        xa: Optional[Tensor] = None,
        mask: Optional[Tensor] = None,
        kv_cache: Optional[dict] = None,
    ):
        x = x + self.attn(self.attn_ln(x), mask=mask, kv_cache=kv_cache)[0]
        if self.cross_attn_ln:
            x = x + self.cross_attn(self.cross_attn_ln(x), xa, kv_cache=kv_cache)[0]
        x = x + self.mlp(self.mlp_ln(x))
        return x


class Model(nn.Module):
    def __init__(
        self,
        n_ctx: int,
        n_state: int,
        n_head: int,
        n_layer: int,
        output_dim: int,
    ) -> None:
        super().__init__()
        self.conv1 = nn.Sequential(
            nn.Conv2d(
                in_channels=1,
                out_channels=n_ctx // 2,
                kernel_size=(3, 7),
                stride=(1, 3),
                padding=(1, 3),
            ),
            nn.ReLU(),
            nn.BatchNorm2d(n_ctx // 2),
        )
        self.pool = nn.AvgPool2d(2, 2)
        self.conv2 = DirectionalConv2d(
            in_channels=n_ctx // 2,
            out_channels=n_ctx,
            kernel=3,
            stride=1,
            padding=1,
            shape=(40, 32),
            extras=[],
        )
        self.register_buffer("positional_embedding", sinusoids(n_ctx, n_state))
        self.blocks = nn.ModuleList(
            [ResidualAttentionBlock(n_state, n_head) for _ in range(n_layer)]
        )
        self.ln_post = nn.LayerNorm(n_state)
        self.collapsing_weights = nn.Parameter(
            torch.nn.init.xavier_uniform_(torch.rand((n_state, 1)))
        )
        self.classification_head = nn.Sequential(
            nn.Linear(n_ctx, n_state),
            nn.ReLU(),
            nn.Dropout(0.4),
            nn.Linear(n_state, output_dim),
        )

    def forward(self, x):
        x = x.unsqueeze(1)
        x = self.conv1(x)
        x = self.pool(x)
        x = self.conv2(x)
        x = self.pool(x)
        x = x.flatten(2)
        assert x.shape[1:] == self.positional_embedding.shape, "incorrect audio shape"
        x = (x + self.positional_embedding).to(x.dtype)
        for block in self.blocks:
            x = block(x)
        x = self.ln_post(x)
        x = x @ self.collapsing_weights
        x = x.flatten(1)
        x = self.classification_head(x)
        return x


def model_lite():
    return Model(n_ctx=256, n_state=320, n_head=4, n_layer=0, output_dim=256)



import numpy as np
from datetime import timedelta
from os.path import basename, isfile


class Segment:
    def __init__(
        self,
        start: float = 0.0,
        end: float = 0.0,
        speaker: str = "",
        confidence: float | None = None,
        filename: str | None = None,
    ) -> None:
        self.start = start
        self.end = end
        self.speaker = speaker
        self.filename = filename
        self.confidence = 1.0 if confidence is None else confidence

    def len(self):
        return self.end - self.start

    def __repr__(self) -> str:
        return f"<Segment({self.start}, {self.end}), speaker {self.speaker}>"

    def __str__(self) -> str:
        return f"{self.pprint(self.start)} -> {self.pprint(self.end)} : {self.speaker if self.speaker is not None else '[silence]'}"

    def pprint(self, t: timedelta, precision: int = 2):
        """
        Return pretty string representation of timedelta
        Arguments:
            - t: timedelta
            - precision: int - number of digits after dot
        """
        return f"{str(timedelta(seconds=t//1))}" + (
            f".{t%1:.2f}"[2 : 2 + precision] if precision > 0 else ""
        )


class Segments:
    """
    Class containing list of Segment instances
    """

    def __init__(self, segments: list[Segment] | None = None) -> None:
        self._data: list[Segment] = [] if segments is None else segments

    def __iter__(self):
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)

    def max(self) -> float:
        if len(self._data) > 0:
            return self._data[-1].end
        else:
            return 0

    def filter(self, speaker):
        out = Segments()
        for el in self._data:
            if el.speaker == speaker:
                out.append(el)
        return out

    def min(self) -> float:
        if len(self._data) > 0:
            return self._data[0].start
        else:
            return 0

    def __getitem__(self, index) -> Segment:
        return self._data[index]

    def __delitem__(self, index) -> None:
        del self._data[index]

    def __add__(self, other):
        if isinstance(other, Segments):
            return Segments(
                segments=self._data + other._data,
            )
        return NotImplemented

    def append(self, el: Segment) -> None:
        self._data.append(el)

    def adjust(self, edges: np.ndarray, tolerance: float):
        """
        Moves both ends of segments towards edges.
        Arguments:
            - edges: np.array of reference edges
            - tolerance: max amount of moving it allowed to
        """
        for i in range(len(self._data)):
            if self._data[i].speaker is None:
                continue
            if (i + 1) < len(self._data) and self._data[i + 1].speaker != self._data[
                i
            ].speaker:
                pred_edge = self._data[i].end
                closest = np.argmin(np.abs(edges - pred_edge))
                dist = self._data[i].end - edges[closest]
                absdist = np.abs(dist)
                if absdist > 0 and absdist < tolerance:
                    self._data[i].end = edges[closest] + (0 if dist <= 0 else -1)
                    self._data[i + 1].start = edges[closest] + (1 if dist <= 0 else 0)

    def merge(self, collapse: bool = False, tolerance: float = 0.2):
        """
        Merges segments with same speaker. If collapse is True then ignores silent intervals between.
        Arguments:
            - collapse: bool [False]
        """
        i = len(self._data)
        while i > 0:
            i -= 1
            for j in reversed(range(i)):
                if self._data[j].speaker == None:
                    if collapse is True:
                        continue
                    else:
                        break
                elif self._data[j].speaker != self._data[i].speaker:
                    break
                else:
                    if self._data[i].start - self._data[j].end <= tolerance:
                        self._data[j].end = self._data[i].end
                        for k in reversed(range(j + 1, i + 1)):
                            del self._data[k]
                    i = j + 1
                    break

    def drop_short(self, min_len: float = 0.1):
        """
        Removes short segments
        Arguments:
            - min_len: min length to remove, in seconds
        """
        for i in reversed(range(len(self._data))):
            if (self._data[i].end - self._data[i].start) <= min_len:
                del self._data[i]

    def drop_silent(self):
        """
        Removes silent|no-speech segments
        """
        for i in reversed(range(len(self._data))):
            if self._data[i].speaker == None:
                del self._data[i]

    def speakers(self) -> list:
        """
        Returns _sorted_ list of _unique_ speakers from segments
        """
        return sorted(list(set([el.speaker for el in self._data])))

    def __repr__(self):
        return f"<Segments [{len(self._data)}]>"

    def __str__(self):
        return "\n".join([str(segment) for segment in self._data])

    def __list__(self):
        return list(self)

    def to_rttm(self, filename: str):
        fname = basename(filename).split(".")[0]
        lines = []
        for seg in self._data:
            lines.append(
                f"SPEAKER {fname} 1 {seg.start:.6f} {(seg.end-seg.start):.6f} <NA> <NA> {seg.speaker} {seg.confidence if seg.confidence is not None else '<NA>'} <NA>\n"
            )
        return lines

    def to_rttm_save(self, fpath: str, overwrite: bool = False):
        if isfile(fpath) and overwrite is False:
            input("File already exists, press Enter to overwrite..")
        lines = self.to_rttm(fpath)
        with open(fpath, "wt") as f:
            f.writelines(lines)

    def from_rttm(self, rttm: str | list):
        """
        Reads data from rttm (could be string or list)
        Arguments:
            - rttm:
        """
        if isinstance(rttm, str):
            rttm = rttm.splitlines()
        for line in rttm:
            split = line.split(" ")
            assert len(split) == 10, "Incorrect number of elements in rttm string"
            filename = split[1]
            onset = float(split[3])
            duration = float(split[4])
            speaker = split[7]
            try:
                confidence = float(split[8])
            except:
                confidence = None
            self._data.append(
                Segment(
                    start=onset,
                    end=onset + duration,
                    speaker=speaker,
                    filename=filename,
                    confidence=confidence,
                )
            )
        return self

    def from_rttm_load(self, fpath: str):
        with open(fpath, "rt") as f:
            self.from_rttm(f.readlines())
        return self

    def to_lab(self):
        lines = []
        for seg in self._data:
            lines.append(f"{seg.start:.6f} {seg.end:.6f} {seg.speaker}\n")
        return lines

    def to_lab_save(self, fpath):
        lines = self.to_lab()
        with open(fpath, "wt") as f:
            f.writelines(lines)

    def from_lab(self, lab: str | list):
        if isinstance(lab, str):
            lab = lab.splitlines()
        for line in lab:
            split = line.split(" ")
            assert len(split) == 3, "Incorrect number of elements in lab string"
            start = float(split[0])
            end = float(split[1])
            speaker = split[2]
            self._data.append(Segment(start=start, end=end, speaker=speaker))

    def from_lab_load(self, fpath):
        with open(fpath, "wt") as f:
            self.from_lab(f.readlines())



import numpy as np
import torch
import librosa
import soxr
import os
from soundfile import read as read_audio
from subprocess import CalledProcessError, run
from silero_vad import get_speech_timestamps, load_silero_vad
from math import ceil


class SampleProcessor:
    def __init__(
        self,
        filename: str,
        use_vad: bool = True,
        sr: int = 16000,
        mono: bool = True,
        top_db: int = 30,
        hop: int = 160,
        n_mels: int = 80,
        n_fft: int = 400,
        window_size: int = 192,
        stride: int = 1,
        slice: int = 60,
        strategy: str = "pad",  # pad or tile
    ):
        self.ratio = sr // hop
        assert self.ratio == sr / hop, "Sample rate should be divisible by hop length"
        self.use_vad = use_vad
        self.sr = sr
        self.mono = mono
        self.top_db = top_db
        self.hop = hop
        self.n_mels = n_mels
        self.n_fft = n_fft
        self.window_size = window_size
        self.stride = stride
        self.strategy = strategy
        self.slice_length = slice * sr // hop
        self.weights = self.compute_weights()
        # Load audio in seconds
        self.audio = self.load_audio(filename)
        self.__spectrogram_full = self.log_mel_spectrogram(audio=self.audio)
        # Get intervals in spectrogram steps
        if use_vad is True:
            self.vad = load_silero_vad()
            intervals = self.detect_voice(self.audio)
            nonsilent = self.get_nonsilent_intervals(audio=self.audio)
            self.__intervals = self.intersect_intervals(intervals, nonsilent)
        else:
            self.__intervals = self.get_nonsilent_intervals(audio=self.audio)
        # Generate mask of non-silent/silent parts of spectrogram[!] in spectrogram steps
        self.mask = self.compute_mask(self.__intervals)
        self.slice_spectrograms()

    def compute_weights(self):
        half = self.window_size // 2
        output = (np.arange(self.window_size) - half) / half
        return np.exp(-(output**2))

    def time_to_step(self, t: float):
        return int(t * self.ratio)

    def step_to_time(self, step: int, mode: str = "floor"):
        if mode == "floor":
            return step / self.ratio
        elif mode == "ceil":
            return (step + 1) / self.ratio
        else:
            raise Exception("Unknown mode")

    def detect_voice(self, audio: np.ndarray):
        raw_intervals = get_speech_timestamps(audio.numpy(), self.vad)
        intervals = np.array(
            [
                [
                    np.floor(el["start"] / self.hop),
                    np.minimum(
                        np.ceil(el["end"] / self.hop),
                        self.__spectrogram_full.size(1) - 1,
                    ),
                ]
                for el in raw_intervals
            ]
        ).astype(int)
        return intervals

    def slice_spectrograms(self):
        # Split into smaller spectrograms (silence removed), prefer to divide using silence
        lengths = [el[1] - el[0] + 1 for el in self.__intervals]
        self.__sequences = []
        summae, seq = 0, []
        for i in range(len(lengths)):
            summae += lengths[i]
            seq += [i]
            if summae >= self.slice_length:
                self.__sequences.append(seq)
                summae, seq = 0, []
        else:
            if summae > 0:
                self.__sequences.append(seq)
        assert np.sum(np.arange(len(lengths))) == sum(
            [sum(el) for el in self.__sequences]
        ), "Missing elements in sequences"
        # Compute sums
        grouped_sums = [
            np.cumsum([el[1] - el[0] + 1 for el in self.__intervals[seq]])
            for seq in self.__sequences
        ]
        self.__sums = np.array([el for seq in grouped_sums for el in seq])

        self.__spectrograms = [
            self.__spectrogram_full[:, self.compute_mask(self.__intervals[seq])]
            for seq in self.__sequences
        ]
        for i, spg in enumerate(self.__spectrograms):
            if spg.shape[1] < self.window_size:
                if self.strategy == "pad":
                    self.__spectrograms[i] = self.pad_spectrogram(spg)
                elif self.strategy == "tile":
                    self.__spectrograms[i] = self.tile_spectrogram(spg)
                else:
                    raise Exception(f"Unknown spectrogram strategy: {self.strategy}")

    def get_spg_pos(self, pos: int):
        """
        returns index of spectrogram and position in spectrogram steps by position in audio
        returns None if position is located in silent part
        """
        condition = (self.__intervals[:, 0] <= pos) & (self.__intervals[:, 1] >= pos)
        index = np.where(condition)[0]
        if len(index) == 0:
            return None
        i = index[0]
        spgpos = self.__sums[i] - (self.__intervals[i, 1] - pos) - 1
        return [i in el for el in self.__sequences].index(True), spgpos

    def get_pos_by_spg_index(self, index: int):
        return (
            self.__intervals[self.__sequences[index][0]][0],
            self.__intervals[self.__sequences[index][-1]][1],
        )

    def get_spectrogram(self, index: int):
        """
        returns spectrogram, silence removed
        """
        return self.__spectrograms[index]

    def get_sliding_windows(self, index: int):
        return (
            self.get_spectrogram(index)
            .unfold(1, self.window_size, self.stride)
            .transpose(1, 0)
        )

    def get_windows_by_spgpos(self, index: int, pos: int):
        """
        returns windows by position in spectrogram steps
        if they were generated altogether per whole spectrogram
        """

        starts = np.arange(
            0, self.__spectrograms[index].size(1) - self.window_size + 1, self.stride
        )
        indices = np.where((starts <= pos) & (pos < (starts + self.window_size)))[0]
        weights = self.weights[pos - starts[indices]]
        return indices, weights

    def sliding_window_indices(self, index: int):
        """
        returns windows by interval in spectrogram steps
        if they were generated altogether per whole spectrogram
        """
        arr_len = self.get_spectrogram(index).size(1)
        # Calculate the total number of sliding windows
        num_windows = (arr_len - self.window_size) // self.stride + 1

        # Create an array of start indices for each sliding window
        start_indices = np.arange(num_windows) * self.stride

        # Create a matrix where each row is a range of indices for the sliding window
        windows = start_indices[:, None] + np.arange(self.window_size)

        # Create an array to store the indices for each element in the original array
        element_to_windows = np.full((arr_len, num_windows), False)

        # For each sliding window, mark the corresponding elements in the array
        for i in range(self.window_size):
            element_to_windows[windows[:, i], np.arange(num_windows)] = True

        # Convert the boolean matrix to a list of indices
        indices = [np.where(element_to_windows[i])[0] for i in range(arr_len)]
        # positions = [np.full_like(ix, i) - ix for i, ix in enumerate(indices)]
        weights = [
            self.weights[np.full_like(ix, i) - ix * self.stride]
            for i, ix in enumerate(indices)
        ]
        return indices, weights

    def load_audio(self, filename: str) -> torch.Tensor:
        """
        Soundfile/Librosa-based audio loader
        """
        try:
            data, sample_rate = read_audio(filename)
        except:
            # fallback to whisper's ffmpeg loader
            data = self.load_audio_whisper(filename, self.sr)
            sample_rate = self.sr
        if self.mono is True and len(data.shape) > 1:
            data = data.mean(axis=1)
        if self.sr != sample_rate:
            data = np.apply_along_axis(
                soxr.resample,
                axis=-1,
                arr=data,
                in_rate=sample_rate,
                out_rate=self.sr,
                quality="soxr_hq",
            )
        assert (
            len(data) > self.window_size * self.hop
        ), "File is too short, expected at least %.2fs" % (
            self.window_size * self.hop / self.sr
        )

        return torch.from_numpy(data)

    @staticmethod
    def load_audio_whisper(file: str, sr: int = 16000):
        """
        Open an audio file and read as mono waveform, resampling as necessary, borrowed from whisper

        Parameters
        ----------
        file: str
            The audio file to open

        sr: int
            The sample rate to resample the audio if necessary

        Returns
        -------
        A NumPy array containing the audio waveform, in float32 dtype.
        """

        # This launches a subprocess to decode audio while down-mixing
        # and resampling as necessary.  Requires the ffmpeg CLI in PATH.
        # fmt: off
        cmd = [
            "ffmpeg",
            "-nostdin",
            "-threads", "0",
            "-i", file,
            "-f", "s16le",
            "-ac", "1",
            "-acodec", "pcm_s16le",
            "-ar", str(sr),
            "-"
        ]
        # fmt: on
        try:
            out = run(cmd, capture_output=True, check=True).stdout
        except CalledProcessError as e:
            raise RuntimeError("Failed to load audio:" + str(e.stderr.decode())) from e

        return np.frombuffer(out, np.int16).flatten().astype(np.float32) / 32768.0

    def log_mel_spectrogram(self, audio: torch.Tensor) -> torch.Tensor:
        """
        Computes the log-Mel spectrogram of

        Parameters
        ----------
        audio: Union[str, np.ndarray, torch.Tensor], shape = (*)
            The path to audio or either a NumPy array or Tensor containing the audio waveform in 16 kHz

        n_mels: int
            The number of Mel-frequency filters, only 80 is supported right now

        Returns
        -------
        torch.Tensor, shape = (n_mels, n_frames)
            A Tensor that contains the Mel spectrogram
        """

        def mel_filters(device, n_mels: int) -> torch.Tensor:
            """
            load the mel filterbank matrix for projecting STFT into a Mel spectrogram.
            Allows decoupling librosa dependency; saved using:

                np.savez_compressed(
                    "mel_filters.npz",
                    mel_64=librosa.filters.mel(sr=16000, n_fft=400, n_mels=64),
                    mel_80=librosa.filters.mel(sr=16000, n_fft=400, n_mels=80),
                    mel_128=librosa.filters.mel(sr=16000, n_fft=400, n_mels=128),
                )
            """
            assert n_mels in {64, 80, 128}, "Unsupported n_mels:" + str(n_mels)

            filters_path = os.path.join(
                os.path.dirname(__file__), "assets", "mel_filters.npz"
            )
            with np.load(filters_path, allow_pickle=False) as f:
                return torch.from_numpy(f[f"mel_{n_mels}"]).to(device)

        window = torch.hann_window(self.n_fft).to(audio.device)
        stft = torch.stft(
            audio, self.n_fft, self.hop, window=window, return_complex=True
        )
        magnitudes = stft[..., :-1].abs() ** 2

        filters = mel_filters(audio.device, self.n_mels).to(audio.dtype)
        mel_spec = filters @ magnitudes

        log_spec = torch.clamp(mel_spec, min=1e-10).log10()
        log_spec = torch.maximum(log_spec, log_spec.max() - 8.0)
        log_spec = (log_spec + 4.0) / 4.0
        return log_spec

    def pad_spectrogram(self, array: torch.Tensor) -> torch.Tensor:
        assert array.shape[0] == self.n_mels, "Incorrect number of mels"
        if array.shape[1] < self.window_size:
            diff = self.window_size - array.shape[1]
            pw = diff // 2
            array = torch.nn.functional.pad(array, (pw, diff - pw), "constant", 0)
        return array

    def tile_spectrogram(self, array: torch.Tensor) -> torch.Tensor:
        assert array.shape[0] == self.n_mels, "Incorrect number of mels"
        if array.shape[1] < self.window_size:
            output = array.repeat((1, ceil(self.window_size / array.shape[1])))[
                :, : self.window_size
            ]
            return output
        return array

    def compute_mask(
        self,
        intervals: np.ndarray,
    ):
        """
        Computes boolean mask of silence measured in spectrogram steps,
        True for non-silent, False for silent parts
        """
        mask = np.zeros(self.__spectrogram_full.size(1), dtype=bool)
        mask[np.concatenate([np.arange(el[0], el[1] + 1) for el in intervals])] = 1
        return mask

    def get_nonsilent_intervals(
        self, audio: torch.Tensor, tolerance: int = 10
    ) -> np.ndarray:
        """
        Returns non-silent intervals in spectrogram steps based on original audio
        """
        intervals = librosa.effects.split(y=audio.numpy(), top_db=self.top_db)
        spgint = np.zeros_like(intervals)
        spgint[:, 0] = np.floor(intervals[:, 0] / self.hop)
        spgint[:, 1] = np.minimum(
            np.ceil(intervals[:, 1] / self.hop), self.__spectrogram_full.size(1) - 1
        )
        # remove too short silent pauses
        removables = []
        for i in reversed(range(1, len(spgint))):
            if (spgint[i, 0] - spgint[i - 1, 1]) < tolerance:
                spgint[i - 1, 1] = spgint[i, 1]
                removables.append(i)
        return np.delete(spgint, removables, 0)

    def intersect_intervals(self, voice: np.ndarray, nonsilent: np.ndarray):
        """
        In addition to VAD also removes parts considered 'silent' (30db difference from peak)
        """
        mask = self.compute_mask(voice)
        silence_mask = ~self.compute_mask(nonsilent)
        mask[silence_mask] = False
        return self.mask_to_intervals(mask)

    def mask_to_intervals(self, mask: np.ndarray) -> np.ndarray:
        change_points = np.diff(mask.astype(int), prepend=0, append=0)
        start_points = np.where(change_points == 1)[0]
        end_points = np.where(change_points == -1)[0] - 1
        intervals = list(zip(start_points, end_points))
        return np.array(intervals)

    def get_fullspectrogram_size(self):
        return self.__spectrogram_full.size(1)

    def get_spectrogram(self, index: int):
        return self.__spectrograms[index]

    def count_spectrograms(self):
        return len(self.__spectrograms)

    def get_intervals(self):
        return self.__intervals

    def iter_intervals(self, seq_index):
        """
        Generates iterator over spectrogram[seq_index].
        Returns intervals of full spectrogram, related to slice spectrogram[index] plus
        locations of intervals within spectrogram slice
        """
        indices = self.__sequences[seq_index]

        def ci(intervals: np.ndarray, indices: list, sums: list):
            s = np.concatenate(([0], sums[indices]))
            for num, i in enumerate(indices):
                yield intervals[i], np.array([s[num], s[num + 1] - 1])

        return ci(intervals=self.__intervals, indices=indices, sums=self.__sums)
